"""
The MoaT Modbus client and its sub-objects (excluding individual bus values).
"""

from __future__ import annotations

import logging
import socket
import struct
from contextlib import asynccontextmanager, suppress
from functools import partial
from pathlib import Path
from typing import Any

import anyio
from anyio import ClosedResourceError, IncompleteRead
from anyio.abc import SocketAttribute
from anyio_serial import Serial
from moat.util import CtxObj, Queue, ValueEvent, num2id
from moat.util.exc import ungroup
from pymodbus.exceptions import ModbusIOException
from pymodbus.pdu import DecodePDU, ExceptionResponse
from pymodbus.framer import FramerRTU, FramerSocket

from .types import BaseValue, DataBlock, TypeCodec, MAX_REQ_LEN

_logger = logging.getLogger(__name__)

__all__ = [
    "ModbusClient",
    "ModbusError",
]

# seconds to wait when disconnecting/reconnecting TCP for different unitID
DISCONNECT_DELAY = 0.1
RECONNECT_TIMEOUT = 10
CHECK_STREAM_TIMEOUT = 0.001


class ModbusClient(CtxObj):
    """The main bus handler. Use as
    >>> async with ModbusClient() as bus:
        ...
    """

    _tg = None

    def __init__(self):
        self.hosts = {}

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as tg:
            self._tg = tg
            try:
                yield self
            finally:
                tg.cancel_scope.cancel()
                self._tg = None
                self.hosts = {}

    def host(self, addr, port=None, **kw):
        """Return a host object for connections to this address+port.

        You cannot create two host objects for the same destination.
        """
        if not port:
            port = 502
        if (addr, port) in self.hosts:
            raise KeyError(f"Host {addr}:{port} already exists")

        h = Host(self, addr, port, **kw)
        return h

    async def _host(self, addr, port=None, *, task_status):
        async with self.host(addr, port) as srv:
            task_status.started(srv)
            await anyio.sleep_forever()

    async def host_service(self, addr, port):
        """Run a TCP client in an AsyncScope."""
        if not port:
            port = 502
        #return await scope.service(f"MC_{num2id(self)}:{addr}:{port}", self._host, addr, port)
        return await self._tg.start(self._host,addr,port)

    def serial(self, /, port, **ser):
        """Return a host object for connections to this serial port."""
        if port in self.hosts:
            raise KeyError(f"Host {port} already exists")

        h = SerialHost(self, port=port, **ser)
        return h

    async def _serial(self, /, port, ser, *, task_status):
        async with self.serial(port, **ser) as srv:
            task_status.started(srv)
            await anyio.sleep_forever()

    async def serial_service(self, port, **ser):
        """Run a serial client in an AsyncScope."""
        # return await scope.service(f"MC_{num2id(self)}:{port}", self._serial, port, **ser)
        return await self._tg.start(self._serial, port, ser)

    def conn(self, cfg):
        """Run a serial OR TCP client according to the config.

        @cfg is a dict with either host/port or port/serial keys.
        Returns a host.

        Usage::
            async with ModbusClient() as g, g.conn(cfg) as c, c.unit(cfg["unit"]) as u:
                ...
        """
        kw = {}
        for k in ("max_rd_len", "max_wr_len"):
            if k in cfg:
                kw[k] = cfg[k]

        if "serial" in cfg:
            port = cfg.get("port", None)
            kw.update(cfg["serial"])
            if port is not None:
                kw["port"] = port
            return self.serial(**kw)

        elif "host" in cfg or ("port" in cfg and isinstance(cfg["port"], int)):
            for k in ("host", "port"):
                with suppress(KeyError):
                    kw[k] = cfg[k]
            return self.host(**kw)
        else:
            raise ValueError("neither serial nor TCP config found")


class ModbusError(RuntimeError):
    """Error entry in returned datasets"""

    def __init__(self, result):
        super().__init__()
        self.result = result


class _HostCommon:
    stream = None
    framer = None  # overridden
    _trace = lambda *x: None  # pylint:disable=unnecessary-lambda-assignment  #  overridden

    def __init__(self, gate, timeout, cap):
        self.gate = gate
        self.units = {}
        self._wqueue = Queue(100)
        self._transactions = {}
        self._tid = 0
        self._tg = None
        self._read_scope = None
        self._connected = anyio.Event()

        self._send_lock = anyio.Lock()

        self.cap = anyio.CapacityLimiter(cap)
        self.timeout = timeout

    def unit(self, unit):
        """
        Returns the `Unit` object registered to nr. @unit.

        A new unit is allocated if it doesn't yet exist.
        """
        return Unit(self, unit)

    async def _unit(self, unit, *, task_status):
        async with self.unit(unit) as srv:
            task_status.started(srv)
            await anyio.sleep_forever()

    async def unit_scope(self, unit):
        """Run a unit in an `AsyncScope`"""
        # return await scope.service(f"MH_{num2id(self)}:{unit}", self._unit, unit)
        return await self._tg.start(self._unit, unit)

    @property
    def _gate_key(self):
        raise NotImplementedError

    @asynccontextmanager
    async def _ctx(self):
        # might be used to manage the connection
        key = self._gate_key
        if key in self.gate.hosts:
            raise RuntimeError(f"Host {key} already exists")
        self.gate.hosts[key] = self

        try:
            async with anyio.create_task_group() as tg:
                self._tg = tg
                self._read_scope = tg.cancel_scope
                await tg.start(self._reader)

                yield self
                tg.cancel_scope.cancel()
        finally:
            self._tg = None
            if self.gate.hosts.get(key, None) is self:
                del self.gate.hosts[key]

    async def send(self, msg):
        """
        Send a packet.

        This is a low-level function. If you are a client, you want to use
        `execute` instead.
        """
        packet = self.framer.buildFrame(msg)
        await self._connected.wait()
        async with self._send_lock:
            await self.stream.send(packet)

    async def execute(self, request):
        """
        Send a pymodbus request and wait for / return the reply.
        """
        # pylint: disable=logging-fstring-interpolation,protected-access

        request.transaction_id = self._nextTID()
        packet = self.framer.buildFrame(request)

        packet_info = " ".join([hex(x) for x in packet])
        self._trace("Gateway xmit: %s", packet_info)

        # make the modbus request
        async with self.cap:
            request._response_value = ValueEvent()

            await self._connected.wait()

            try:
                self._transactions[request.transaction_id] = request
                await self.send(request)
                with anyio.fail_after(self.timeout):
                    res = await request._response_value.get()

                if res.isError():
                    raise ModbusError(res)

                if hasattr(res, "registers"):
                    registers_info = " ".join([hex(x) for x in res.registers])
                    self._trace("Gateway replied: %s", registers_info)
                else:
                    self._trace("Gateway replied: %s", res)
                return res

            finally:
                self._transactions.pop(request.transaction_id, None)


class Host(_HostCommon, CtxObj):
    """This is a single host which moat-modbus talks to.
    It has a number of modbus units (attribute 'units').

    Do not instantiate directly; instead, use

        >>> async with client.host("foo.example" [, port=20502] ) as host:
            ...
    """

    _tg = None

    def __init__(
        self,
        gate,
        addr,
        port,
        timeout=10,
        cap=1,
        debug=False,
        max_rd_len=MAX_REQ_LEN,
        max_wr_len=MAX_REQ_LEN,
    ):
        self.addr = addr
        self.port = port

        self.max_rd_len = max_rd_len
        self.max_wr_len = max_wr_len

        log = logging.getLogger(f"modbus.{addr}")
        self._trace = log.info if debug else log.debug

        self.framer = FramerSocket(DecodePDU(False))

        super().__init__(gate, timeout, cap)

    def __repr__(self):
        return f"<ModbusHost:{self.addr}:{self.port}>"

    def _nextTID(self):
        self._tid = (self._tid + 1) % 0xFFFF
        return self._tid

    @property
    def _gate_key(self):
        key = (self.addr, self.port)

    # reader task #

    async def _reader(self, *, task_status):
        # pylint: disable=protected-access

        async def _send_trans():
            tr = list(self._transactions.values())
            self._transactions = {}
            if tr:
                _logger.warning("Resend %d packets", len(tr))
            try:
                for request in tr:
                    packet = self.framer.buildFrame(request)
                    async with self._send_lock:
                        await self.stream.send(packet)
                if tr:
                    _logger.warning("Resend done")
            except Exception:  # pylint: disable=broad-except
                _logger.exception("Re-Write")

        data = bytearray()

        while True:
            try:
                if self.stream is None:
                    with anyio.fail_after(self.timeout):
                        self.stream = await anyio.connect_tcp(self.addr, self.port)
                        # set so_linger to force sending RST instead of FIN
                        self.stream.extra(SocketAttribute.raw_socket).setsockopt(
                            socket.SOL_SOCKET,
                            socket.SO_LINGER,
                            struct.pack("ii", 1, 0),
                        )
                        # re-send open requests
                        await _send_trans()
                        self._connected.set()
                        if task_status is not None:
                            task_status.started()
                            task_status = None
                    data = bytearray()

                data += await self.stream.receive(4096)
                # pylint: disable=logging-not-lazy
                self._trace("recv: " + " ".join([hex(x) for x in data]))

                replies = []

                while True:
                    used, pdu = self.framer.processIncomingFrame(data)
                    data = data[used:]
                    if pdu is not None:
                        replies.append(pdu)
                    if not used:
                        break

            except (
                IncompleteRead,
                ConnectionRefusedError,
                ConnectionResetError,
                ClosedResourceError,
                ModbusIOException,
                anyio.BrokenResourceError,
                anyio.EndOfStream,
                TimeoutError,
            ) as exc:
                if self._connected.is_set():
                    self._connected = anyio.Event()
                t, self._transactions = self._transactions, {}
                if t:
                    for req in t.values():
                        req._response_value.set_error(exc)
                else:
                    _logger.error(
                        "Read from %s:%d: %r (%d)",
                        self.addr,
                        self.port,
                        exc,
                        len(self._transactions),
                    )

                s, self.stream = self.stream, None
                if s:
                    await s.aclose()

                # delay somewhat, to give the device the chance to reinitialize
                await anyio.sleep(DISCONNECT_DELAY)

            except anyio.get_cancelled_exc_class():
                raise

            except BaseException as exc:
                _logger.exception("Error: %r", exc)

                t, self._transactions = self._transactions, {}
                for req in t.values():
                    req._response_value.set_error(exc)
                raise

            else:
                for reply in replies:
                    tid = reply.transaction_id
                    try:
                        request = self._transactions.pop(tid)
                    except KeyError:
                        _logger.info("Unrequested message: %s", reply)
                    else:
                        request._response_value.set(reply)

    async def aclose(self):
        """Stop talking."""
        if self.gate is None:
            return
        self.gate._del_host(self)  # pylint: disable=protected-access
        self.gate = None

        u, self.units = self.units, None
        for unit in u.values():
            await unit.aclose()

        if self._read_scope is not None:
            self._read_scope.cancel()
            self._read_scope = None

        s, self.stream = self.stream, None
        if s:
            await s.close()


class SerialHost(_HostCommon, CtxObj):
    """This is a "host" that's actually a serial interface.

    Do not instantiate directly; instead, use

        >>> async with client.serial("/dev/ttyUSB0",
                baudrate=9600, parity="E", stopbits=1) as host:
            ...
    """

    _tg = None

    def __init__(
        self,
        gate,
        /,
        port,
        timeout=10,
        debug=False,
        monitor=None,
        max_rd_len=MAX_REQ_LEN,
        max_wr_len=MAX_REQ_LEN,
        **ser,
    ):
        self.port = port
        self.ser = ser
        self.framer = FramerRTU(DecodePDU(False))
        self.max_rd_len = max_rd_len
        self.max_wr_len = max_wr_len

        log = logging.getLogger(f"modbus.{Path(port).name}")
        self._trace = log.info if debug else log.debug
        self._monitor = monitor

        super().__init__(gate, timeout, 1)

    def __repr__(self):
        return f"<ModbusHost:{self.port}:{self.ser.get('baudrate', 0)}>"

    def _nextTID(self):
        return 0

    @property
    def _gate_key(self):
        key = self.port

    # reader task #

    async def _reader(self, *, task_status):
        # pylint: disable=protected-access

        if self._transactions:
            raise RuntimeError("Serial: cannot have open transaction on start")
        task_status.started()
        self._trace("recv START")

        mon = self._monitor
        data = bytearray()
        while True:
            try:
                async with ungroup, Serial(port=self.port, **self.ser) as self.stream:
                    self._connected.set()
                    while True:

                        if data:
                            with anyio.fail_after(self.timeout):
                                data += await self.stream.receive(4096)
                        else:
                            data = await self.stream.receive(4096)

                        # pylint: disable=logging-not-lazy
                        self._trace("recv: " + " ".join([hex(x) for x in data]))

                        replies = []

                        # check for decoding errors
                        while True:
                            used, pdu = self.framer.processIncomingFrame(data)
                            if pdu is not None:
                                replies.append(pdu)
                            if not used:
                                break
                            data = data[used:]

                        if mon:
                            for reply in replies:
                                await mon(reply)
                        else:
                            for reply in replies:
                                tid = reply.transaction_id
                                try:
                                    request = self._transactions.pop(tid)
                                except KeyError:
                                    _logger.info("Unrequested message: %s", reply)
                                else:
                                    request._response_value.set(reply)

            except (
                IncompleteRead,
                ConnectionRefusedError,
                ConnectionResetError,
                ClosedResourceError,
                ModbusIOException,
                anyio.BrokenResourceError,
                anyio.EndOfStream,
                TimeoutError,
            ) as exc:
                if self._connected.is_set():
                    self._connected = anyio.Event()
                _logger.error("Read: %r (%d)", exc, len(self._transactions))

                t, self._transactions = self._transactions, {}
                for req in t.values():
                    req._response_value.set_error(exc)

            except anyio.get_cancelled_exc_class():
                raise

            except BaseException as exc:
                _logger.exception("Error: %r", exc)

                t, self._transactions = self._transactions, {}
                for req in t.values():
                    req._response_value.set_error(exc)
                raise


    async def aclose(self):
        """Stop talking."""
        if self.gate is None:
            return
        self.gate._del_host(self)  # pylint: disable=protected-access
        self.gate = None

        u, self.units = self.units, None
        for unit in u.values():
            await unit.aclose()

        if self._read_scope is not None:
            self._read_scope.cancel()
            self._read_scope = None

        s, self.stream = self.stream, None
        if s:
            await s.aclose()


class Unit(CtxObj):
    """This is a single modbus unit. It has any number of time slots
    (attribute 'slots'). Simply indexing a unit will return a timeslot
    object (existing or new). You can use anything hashable as the index.

    Units are always linked to a host. Use

        >>> async with host.unit(1) as u:
            ...

    to access/create a unit.
    """

    _running = False

    def __init__(self, host, unit):
        self.host = host
        self.unit = unit
        self.slots = {}

    def __str__(self):
        try:
            return f"{self.host.addr}:{self.host.port}:{self.unit}"
        except AttributeError:
            return f"{self.host.port}:{self.unit}"

    def __repr__(self):
        try:
            return f"<Unit:{self.host.addr}:{self.host.port}:{self.unit}>"
        except AttributeError:
            return f"<Unit:{self.host.port}:{self.unit}>"

    @asynccontextmanager
    async def _ctx(self):
        if self._running:
            raise RuntimeError(f"Slot {self.slot} already active")
        self._running = True
        try:
            async with anyio.create_task_group() as tg:
                self._tg = tg
                yield self
                tg.cancel_scope.cancel()
        finally:
            self._tg = None
            self._running = False
            self.host.units.pop(self.unit, None)


    def slot(self, slot, **kw):
        """
        Returns the `Slot` object registered to @slot.

        A new slot is allocated if it doesn't yet exist.
        """
        try:
            return self.slots[slot]
        except KeyError:
            self.slots[slot] = sl = Slot(self, slot, **kw)
            return sl

    async def _slot(self, slot, kw, *, task_status):
        async with self.slot(slot, **kw) as srv:
            task_status.started(srv)
            await anyio.sleep_forever()

    async def slot_scope(self, slot, **kw):
        """Run the slot handler in an AsyncScope."""
        # return await scope.service(f"MS_{num2id(self)}:{slot}", self._slot, slot, **kw)
        return await self._tg.start(self._slot, slot, kw)

    async def aclose(self):
        """Stop talking and delete yourself.

        This is a coroutine.
        """
        if self.host is None:
            return
        self.host._del_unit(self)  # pylint: disable=protected-access
        self.host = None

        s, self.slots = self.slots, None
        for slot in s.values():
            await slot.aclose()

    async def process_request(self, request):
        """
        Forward a request to the host.
        """
        request.unit_id = self.unit
        try:
            response = await self.host.execute(request)
        except Exception as exc:
            logger.exception("Handler for %d: %r", unit_id, exc)
            response = ExceptionResponse(request.function_code, ExceptionResponse.SLAVE_FAILURE)

        return response


class Slot(CtxObj):
    """This class represents a single "atomic" access to Modbus. The system
    will periodically try to fetch all values in this slot, using as
    few+small requests as possible.

    Slots are always linked to a unit. Use

        >>> async with unit.slot("20seconds") as slot:
            ...

    to access/create a slot.

    The intended usecase is that some values should be retrieved at
    different intervals than others. Thus the user creates several slots
    and adds the required fields to them.

    Slots will periodically read registers,

    Slots will write updated holding registers, delaying for @write_delay
    for possible collation.
    """

    _get_scope = None
    _set_scope = None
    delay: float = None
    t_read = None
    _scope = None
    _running = False

    def __init__(
        self,
        unit,
        slot,
        read_delay: float | None = None,
        read_align: bool = False,
        write_delay: float | None = None,
        **kw,
    ):
        self.unit = unit
        self.slot = slot

        self.write_delay = write_delay
        self.write_lock = anyio.Lock()
        self.write_trigger = anyio.Event()

        self.read_delay = read_delay
        self.read_align = read_align
        self.read_lock = anyio.Lock()
        self.read_trigger = anyio.Event()

        self.run_lock: anyio.Event = None

        self.modes = {}
        if kw:
            _logger.warning("%s:%s: extra arguments: %r", unit, slot, kw)

    def __str__(self):
        return f"{self.unit}:{self.slot}"

    def __repr__(self):
        return f"<Unit:{self.unit}:{self.slot}>"

    @asynccontextmanager
    async def _ctx(self):
        if self._running:
            raise RuntimeError(f"Slot {self.slot} already active")
        self._running = True
        try:
            async with anyio.create_task_group() as tg:
                self._scope = tg.cancel_scope
                self.run_lock = anyio.Event()
                if self.write_delay is not None:
                    tg.start_soon(self.write_task)
                if self.read_delay is not None:
                    tg.start_soon(self.read_task)
                yield self
                tg.cancel_scope.cancel()
        finally:
            self.run_lock = None
            self._running = False
            del self.unit.slots[self.slot]

    def start(self):
        """
        Start running this slot.
        """
        self.run_lock.set()

    def trigger_send(self):
        """Start writing after at most `.write_delay` seconds."""
        self.write_trigger.set()

    @property
    def is_empty(self):
        """
        Check whether the slot does not contain any registers.
        """
        return all(not offsets.values for offsets in self.modes.values())

    def add(self, typ: TypeCodec, offset: int, cls: type[BaseValue] | BaseValue) -> BaseValue:
        """Add a field to this slot.

        :param typ: The `TypeCodec` instance to use.
        :param offset: The value's numeric offset, zero-based.
        :param val: The data type (baseValue instance)

        `cls` is the decoder (subclass of `BaseValue`).
        """
        try:
            k = self.modes[typ]
        except KeyError:
            self.modes[typ] = k = ValueList(self, typ)
        val = cls(offset=offset) if isinstance(cls, type) else cls
        k.add(offset, val)
        return val

    def remove(self, typ: TypeCodec, offset: int):
        """Remove a field to be requested.

        :param typ: the `TypeCodec` to use
        :param offset: the offset where the value is located

        Returns the field in question, or none if it doesn't exist.
        """
        try:
            k = self.modes[typ]
        except KeyError:
            return None
        return k.delete(offset)

    async def _stop_run(self):
        if self._get_scope is not None:
            await self._get_scope.cancel()
        if self._set_scope is not None:
            await self._set_scope.cancel()

    def close(self):
        """
        Close this slot.
        """
        if self._scope is None:
            return
        self._scope.cancel()

    async def getValues(self) -> dict[TypeCodec, dict[int, Any]]:
        """
        Send messages reading this slot's values from the bus.
        Returns a (type,(offset,value)) dict-of-dicts.
        On error, stores the error object instead of a dict.

        This method may return a partial result.
        """
        try:
            res = {}

            async def _assign(getter, rr):
                r = await getter()
                rr.update(r)

            async with anyio.create_task_group() as tg:
                if self._get_scope is not None:
                    raise RuntimeError("already running")
                self._get_scope = tg.cancel_scope

                for typ, vl in self.modes.items():
                    res[typ] = r = {}
                    try:
                        tg.start_soon(partial(_assign, vl.readValues, r))
                    except ModbusError as r:
                        res[typ] = r.result

            if self.unit is None:
                raise ClosedResourceError("dropped while running")
            return res
        finally:
            self._get_scope = None

    async def setValues(self, changed=False):
        """
        Send a message writing the values in this block to the bus.
        """
        try:
            async with anyio.create_task_group() as tg:
                if self._set_scope is not None:
                    raise RuntimeError("already running")
                self._set_scope = tg.cancel_scope

                for vl in self.modes.values():
                    tg.start_soon(vl.writeValues, changed)
            if self.unit is None:
                raise ClosedResourceError("dropped while running")
        finally:
            self._set_scope = None

    async def read(self):
        """Read this slot's data."""
        async with self.read_lock:
            self.t_read = anyio.current_time()
            await self.getValues()

    async def read_task(self):
        """A background task for reading Modbus register values.
        We read every .`read_delay` seconds.
        """
        await self.run_lock.wait()

        try:
            await self.read()
        except Exception as exc:  # pylint:disable=broad-except
            _logger.warning("Error %s: %r", self, exc)
        tn = self.t_read + self.read_delay
        if self.read_align:
            tn -= tn % self.read_delay

        backoff = 0

        while True:
            t = anyio.current_time()
            tr = self.t_read
            if t < tn + backoff:
                await anyio.sleep(tn + backoff - t)

            async with self.read_lock:
                if self.t_read != tr:
                    # somebody else has read while we waited for the lock
                    tn = self.t_read + self.read_delay
                    if self.read_align:
                        tn -= tn % self.read_delay
                    continue
                if t < tn:
                    # We slept, above, thus update the current time
                    self.t_read = tn
                    tn += self.read_delay
                else:
                    # We didn't sleep: the last read took too long. reset the timer
                    _logger.info(
                        "Delay for %s: %.1f > %.1f",
                        self,
                        t - tn + self.read_delay,
                        self.read_delay,
                    )

                    self.t_read = t = anyio.current_time()
                    tn = t + self.read_delay
                    if self.read_align:
                        tn -= tn % self.read_delay

                try:
                    await self.getValues()  # already locked
                except Exception as exc:  # pylint:disable=broad-except
                    _logger.warning("Error %s: %r", self, exc)
                    backoff = 1 + backoff * 1.2
                    # TODO re-raise if persistent?
                else:
                    backoff = 0

    async def write(self, changed: bool = True):
        """Write this slot's data.

        If @changed is set, only write changed data.
        """
        async with self.write_lock:
            await self.setValues(changed=changed)

    async def write_task(self):
        """A background task for updating changed Modbus register values"""
        await self.run_lock.wait()
        while True:
            await self.write_trigger.wait()
            await anyio.sleep(1)  # self.write_delay)
            self.write_trigger = anyio.Event()
            try:
                await self.write(changed=True)
            except ModbusError:
                _logger.exception("Write %s", self)
                # TODO examine+record the error


class ValueList(DataBlock):
    """This class holds a list of to-be-accessed values.

    Do not instantiate directly; constructed and returned by

        >>> slot.add(InputRegisters, 2, IntValue)

    """

    def __init__(self, slot, kind):
        super().__init__(
            max_rd_len=slot.unit.host.max_rd_len,
            max_wr_len=slot.unit.host.max_wr_len,
        )
        self.slot = slot
        assert hasattr(kind, "encoder_m")
        self.kind = kind
        self.do_write = anyio.Event()

    def trigger_send(self):
        self.slot.trigger_send()

    async def readValues(self, *, res=None):
        """
        Send messages reading these values to the bus.

        Returns a (offset,value) dict.
        """
        if res is None:
            res = {}
        async with anyio.create_task_group() as tg:
            for start, length in self.ranges(max_len=self.max_rd_len):
                tg.start_soon(partial(self.readBlock, start, length, res=res))
        return res

    async def writeValues(self, changed=False):
        """
        Send messages writing our values to the bus.
        """
        async with anyio.create_task_group() as tg:
            for start, length in self.ranges(changed=changed, max_len=self.max_wr_len):
                tg.start_soon(partial(self.writeBlock, start, length))

    async def readBlock(self, start, length, *, res=None):
        """Send one message, decode the reply"""
        if res is None:
            res = {}
        u = self.slot.unit
        msg = self.kind.encoder(address=start, count=length, dev_id=u.unit)

        r = await u.host.execute(msg)

        # TODO check R for correct type?

        r = r.registers

        while r:
            try:
                val = self[start]
            except KeyError:
                off = 1
            else:
                off = val.len
                val.decode(r[:off])
                res[start] = val
            start += off
            r = r[off:]

    async def writeBlock(self, start, length):
        """encode one message and send it"""
        u = self.slot.unit
        values = []
        off = start
        res = length
        while res > 0:
            val = self[off]
            values.extend(val.encode())
            off += val.len
            res -= val.len
        if len(values) == 1:
            msg = self.kind.encoder_s(address=start, dev_id=u.unit, registers=values)
        else:
            msg = self.kind.encoder_m(address=start, count=length, dev_id=u.unit, registers=values)

        await u.host.execute(msg)  # pylint: disable=unused-variable
        # raises an error if failed

        self.markSent(start, length)

    def close(self):
        """disable further accesses"""
        if self.slot is None:
            return
        self.slot = None
        self.values = None



"""
Map strings+kinds to modbuys types
"""

from __future__ import annotations

from functools import partial

from moat.modbus.types import (
    BitValue,
    InvBitValue,
    ByteValue,
    Coils,
    DiscreteInputs,
    DoubleValue,
    FloatValue,
    HoldingRegisters,
    InputRegisters,
    IntValue,
    LongValue,
    QuadValue,
    SignedIntValue,
    SignedLongValue,
    SignedQuadValue,
    StringValue,
    SwappedByteValue,
    SwappedDoubleValue,
    SwappedFloatValue,
    SwappedLongValue,
    SwappedQuadValue,
    SwappedSignedLongValue,
    SwappedSignedQuadValue,
    SwappedStringValue,
)

map_type = {
    "raw": IntValue,
    "x": BitValue,
    "X": InvBitValue,
    "u1": IntValue,
    "U1": IntValue,
    "u2": LongValue,
    "U2": SwappedLongValue,
    "u4": QuadValue,
    "U4": SwappedQuadValue,
    "s1": SignedIntValue,
    "S1": SignedIntValue,
    "s2": SignedLongValue,
    "S2": SwappedSignedLongValue,
    "s4": SignedQuadValue,
    "S4": SwappedSignedQuadValue,
    "f2": FloatValue,
    "F2": SwappedFloatValue,
    "f4": DoubleValue,
    "F4": SwappedDoubleValue,
    "c#": StringValue,
    "C#": SwappedStringValue,
    "b#": ByteValue,
    "B#": SwappedByteValue,
}


def get_type(s):
    """Return the type from shortname (like 'u2')"""
    hashkey = f"{s[0]}#"
    if hashkey in map_type:
        return partial(map_type[hashkey], int(s[1:]))
    else:
        return map_type[s]


def get_type2(s, l):
    """Return the type from longname and length (like 'int' '2')"""
    IntMap = [
        [
            {
                1: SignedIntValue,
                2: SignedLongValue,
                4: SignedQuadValue,
            },
            {
                1: IntValue,
                2: LongValue,
                4: QuadValue,
            },
        ],
        [
            {
                1: SignedIntValue,
                2: SwappedSignedLongValue,
                4: SwappedSignedQuadValue,
            },
            {
                1: IntValue,
                2: SwappedLongValue,
                4: SwappedQuadValue,
            },
        ],
    ]
    FloatMap = [
        {
            2: FloatValue,
            4: DoubleValue,
        },
        {
            2: SwappedFloatValue,
            4: SwappedDoubleValue,
        },
    ]
    swapped = False
    unsigned = False

    os = s
    if s[0] == "s" and s != "str":
        s = s[1:]
        swapped = True
    if s[0] == "u":
        s = s[1:]
        unsigned = True
    if s == "int":
        return IntMap[swapped][unsigned][l]

    if not unsigned:
        if s == "float":
            return FloatMap[swapped][l]
        if s == "byte":
            return partial([ByteValue, SwappedByteValue][swapped], length=l)
        if s == "str":
            return partial([ByteValue, SwappedByteValue][swapped], length=l)
        if not swapped and l == 1:
            if s == "bit":
                return BitValue
            if s == "invbit":
                return InvBitValue
    raise KeyError(f"Unknown: {os}:{l}")


map_kind = {"c": Coils, "d": DiscreteInputs, "h": HoldingRegisters, "i": InputRegisters}


def get_kind(s):
    """Return the value kind from name"""
    return map_kind[s[0]]



"""
Types that describe a modbus device, as read from file
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from contextlib import asynccontextmanager
from copy import deepcopy
from pathlib import Path as FSPath

import anyio
from asyncscope import scope
from moat.util import CtxObj, P, Path, attrdict, combine_dict, merge, yload

from moat.modbus.typemap import get_kind, get_type2
from moat.modbus.types import Coils, DiscreteInputs, InputRegisters
from moat.modbus.server import UnitContext
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.modbus.client import ModbusClient, Slot, Unit

logger = logging.getLogger(__name__)


class BadRegisterError(ValueError):
    """Broken register description"""

    pass


class NotARegisterError(ValueError):
    """Not a register at this location"""

    pass


def mark_orig(d):
    if isinstance(d, dict):
        d._is_orig = True
        for k, v in d.items():
            if k != "default":
                mark_orig(v)


def fixup(d, this_file=None, **k):
    """
    See `fixup_`.

    Also marks original data
    """
    mark_orig(d)
    d = fixup_i(d, this_file=this_file)
    return fixup_(d, **k)


def fixup_i(d, this_file=None):
    """
    Run processing instructions: include
    """
    if isinstance(d, Mapping):
        try:
            inc = d.pop("include")
        except KeyError:
            pass
        else:
            if isinstance(inc, tuple):
                inc = list(inc)
            elif not isinstance(inc, list):
                inc = [inc]
            for i, dd in enumerate(inc):
                try:
                    f = _data / dd
                    df = f.open("r")
                except FileNotFoundError:
                    f = this_file.parent / dd
                    df = f.open("r")
                with df:
                    dd = yload(df, attr=True)
                inc[i] = fixup_i(dd, this_file=f)
            inc.reverse()
            d = combine_dict(d, *inc, cls=attrdict)
            d._root = True

    for k, v in d.items() if hasattr(d, "items") else enumerate(d):
        if isinstance(v, (Mapping, list, tuple)):
            d[k] = fixup_i(v)

    return d


def fixup_(
    d,
    root=None,
    path=Path(),
    post=None,
    default=None,
    offset=0,
    do_refs=True,
    apply_default=False,
):
    """
    Run processing instructions: ref, default, repeat
    """
    if root is None or getattr(d, "_root", False):
        root = d
    else:
        pass
    if default is None:
        default = attrdict()

    reps = set()

    if isinstance(d, Mapping):
        try:
            defs = d.pop("default")
        except KeyError:
            pass
        else:
            default = combine_dict(defs, default, cls=attrdict)

        if do_refs:
            try:
                refs = d.pop("ref")
            except KeyError:
                pass
            else:
                if isinstance(refs, str):
                    refs = P(refs)
                if isinstance(refs, Path):
                    refs = [refs]
                for i, p in enumerate(refs):
                    refs[i] = root._get(p)  # pylint: disable=protected-access

                refs.reverse()
                d = combine_dict(d, *refs, cls=attrdict)

        try:
            rep = d.pop("repeat")
        except KeyError:
            rep = None

        if "register" in d or apply_default:
            if "register" in d:
                d.register += offset
            merge(d, default, replace=False)

        # Offset is modified here
        if rep:
            k = rep.get("start", 0)
            n = rep.n
            off = offset
            while n > 0:
                v = combine_dict(d.get(k, attrdict()), deepcopy(rep.data), cls=attrdict)
                d[k] = fixup_(
                    v,
                    root,
                    path / k,
                    default=default,
                    offset=off,
                    do_refs=do_refs,
                    apply_default=getattr(d, "_apply_default", False),
                )
                reps.add(k)

                n -= 1
                k += 1
                off += rep.offset

    for k, v in d.items() if hasattr(d, "items") else enumerate(d):
        if k in reps:
            continue
        if isinstance(v, (Mapping, list, tuple)):
            d[k] = fixup_(
                v,
                root,
                path / k,
                default=default,
                offset=offset,
                do_refs=do_refs,
                apply_default=getattr(d, "_apply_default", False),
            )

    if post is not None:
        d = post(d, path)

    return d


class Register:
    """A single modbus device's register.

    This class duck-types as a moat.modbus.types.BaseValue."""

    last_gen = -1
    block = None

    def __init__(self, d, path, unit=None):
        try:
            self.reg_type = get_kind(d.reg_type)
        except AttributeError:
            self.reg_type = InputRegisters

        try:
            s = d.type
        except AttributeError:
            if self.reg_type is DiscreteInputs or self.reg_type is Coils:
                s = "bit"
            else:
                raise AttributeError(f"No type in {path}") from None
        else:
            if s in {"bit", "invbit"}:
                pass
            elif self.reg_type is DiscreteInputs or self.reg_type is Coils:
                raise RuntimeError(f"Coils/Discretes must be BitValue, at {path}")

        try:
            l = d.len
        except AttributeError:
            if s in {"bit", "invbit", "int", "uint"}:
                l = 1
            elif s == "float":
                l = 2
            else:
                raise BadRegisterError("no length") from None

        self.reg = get_type2(s, l)()

        self.register = d.register

        if "slot" in d and unit is not None:
            slot = unit.slot(d.slot)
            try:
                slot.add(self.reg_type, offset=self.register, cls=self.reg)
            except ValueError:
                raise ValueError("Already known", slot, self.reg_type, self.register) from None
            self.slot = slot
        self.unit = unit
        self.data = d
        self.factor = 10 ** self.data.get("scale", 0) * self.data.get("factor", 1)
        self.offset = self.data.get("offset", 0)
        self.path = path

    async def start(self):
        pass

    def __aiter__(self):
        return self._iter().__aiter__()

    async def _iter(self):
        async for val in self.reg:
            if val is not None:
                if self.factor != 1:
                    val *= self.factor
                if self.offset:
                    val += self.offset
            yield val

    @property
    def value(self):
        """Returns the factor+offset-adjusted value from the bus"""
        val = self.reg.value
        if val is not None:
            if self.factor != 1:
                val *= self.factor
            if self.offset:
                val += self.offset
        return val

    @property
    def value_w(self):
        """Returns the factor+offset-adjusted value from the bus"""
        val = self.reg.value_w
        if val is not None:
            if self.factor != 1:
                val *= self.factor
            if self.offset:
                val += self.offset
        return val

    @value.setter
    def value(self, val):
        """Sets the value that'll be written to the bus.
        Reverses factor+offset-adjustment.
        Should trigger an update.
        """
        if val is not None:
            if self.offset:
                val -= self.offset
            if self.factor != 1:
                val /= self.factor
        self.reg.set(val)

    @property
    def len(self):
        """number of Modbus registers occupied by this value"""
        return self.reg.len

    def encode(self):
        """Encode myself"""
        return self.reg.encode()

    def decode(self, regs: list[int]):
        """Encode registers into self"""
        self.reg.decode(regs)

    @property
    def changed(self):
        """Flag whether this register needs to be updated"""
        return self.reg.changed

    @property
    def gen(self):
        """Generation number for the current value"""
        return self.reg.gen

    def __repr__(self):
        return f"‹{str(self.reg_type)[0].lower()}{self.register} @{self.path}:{self.value}›"

    __str__ = __repr__


_data = FSPath(__file__).parent / "_data"


class BaseDevice:
    """A modbus device.

    The idea is to use the device description file as a template.

    You augment that file with "slot" data, i.e. named intervals,
    plus processing instructions. The instructions may contain slot names.

    For each slot, the system will periodically fetch the data and call a
    given postprocessor for each item, which can e.g. forward the value to
    MQTT.

    @factory is used to create bus registers. Used to augment basic
    registers, e.g. with different storage backends.
    """

    data: attrdict = None
    cfg: attrdict = None
    cfg_path: FSPath = None

    def __init__(self, factory=Register):
        self.factory = factory

    async def load(self, path: str | None = None, data: dict | None = None):
        """Load a device description from @path, augmented by @data"""
        if self.cfg is not None:
            raise RuntimeError("already called")

        if path is None:
            d = attrdict()
        else:
            path = _data / path
            d = yload(path, attr=True)
        if data is not None:
            d = merge(d, data)
        self.cfg = d
        self.cfg_path = path

    def get(self, path: Path):
        """Fetch the register at a subpath.

        Raises an error if there's no register there.
        """
        dev = self.data._get(path)  # pylint: disable=protected-access
        if not isinstance(dev, Register):
            raise NotARegisterError(path)
        return dev


class ClientDevice(CtxObj, BaseDevice):
    """
    A client device, i.e. one that mirrors some Modbus master's unit
    """

    unit: Unit = None

    def __init__(self, client: ModbusClient, factory=Register):
        super().__init__(factory)
        self.client = client

    @asynccontextmanager
    async def _ctx(self):
        if "host" in self.cfg.src:
            host = await self.client.host_service(self.cfg.src.host, self.cfg.src.get("port"))
        else:
            host = await self.client.serial_service(
                port=self.cfg.src.port,
                **self.cfg.src.get("serial", {}),
            )
        self.unit = await host.unit_scope(self.cfg.src.unit)

        self.data = fixup(self.cfg, root=self.cfg, path=Path(), this_file=self.cfg_path)
        await self.add_slots()
        await self.add_registers()
        yield self

    async def as_scope(self):
        "Basic scope generator"
        async with self:
            scope.register(self)
            await scope.no_more_dependents()

    async def add_slots(self):
        return

    async def add_slots(self):
        """Add configured slots to this instance"""
        if "slots" not in self.data:
            logger.warning("No slots in %r", self)
            return
        for k, v in self.data.slots.items():
            if v is None:
                v = {}
            self.slots[k] = await self.unit.slot_scope(k, **v)

    async def add_registers(self):
        """Replace entries w/ register/slot members with Register instances"""

        async def a_r(d, path=Path()):
            seen = False
            for k, v in d.items():
                if not isinstance(v, dict):
                    continue
                if await a_r(v, path / k):
                    seen = True
                    if "register" in v:
                        logger.warning("%s has a sub-register: ignored", path / k)
                    elif "slot" in v:
                        logger.warning("%s is not a register", path / k)
                    continue

                if "register" in v:
                    v.setdefault("slot", "write")
                    d[k] = f = self.factory(v, path / k, self.unit)
                    await f.start()
                    seen = True
                elif "slot" in v:
                    logger.warning("%s is not a register", path / k)
            return seen

        await a_r(self.data)

    @property
    def slots(self):
        """The slots of this unit"""
        return self.unit.slots

    async def update(self, slot: Slot, proc=None):
        """Update a slot. Calls @proc with each register (in parallel)."""
        vals = await slot.read()
        if proc is not None:
            async with anyio.create_task_group() as tg:
                for d in vals.values():
                    for v in d.values():
                        tg.start_soon(proc, v)
        return vals

    async def poll_slot(self, slot: str, *, task_status=None):
        """Task to register and periodically poll a given slot"""
        # slots:
        #  1sec:
        #    time: 1
        #    align: false
        #
        # align=True: wait for the next multiple
        # align=False: fetch now, *then* wait for the next multiple
        # align=None: fetch now, wait for the timespan

        if task_status is not None:
            task_status.started()

        sl = self.unit.slots[slot]
        sl.start()

        while True:
            await anyio.sleep(99999)

    async def poll(self, slots: set | None = None, *, task_status=None):
        """Task to periodically poll all slots"""
        if slots is None:
            slots = self.data.slots.keys()
        for slot in slots:
            self.unit.slots[slot].start()
        if task_status is not None:
            task_status.started()

        while True:
            await anyio.sleep(99999)


class ServerDevice(BaseDevice):
    """
    A server device, i.e. a unit that's accessed via modbus.
    """

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.unit = UnitContext()

        # duck-type me
        try:
            self.validate = self.unit.validate
        except AttributeError:
            pass
        self.getValues = self.unit.getValues
        self.setValues = self.unit.setValues

    def async_getValues(self,*a,**kw):
        return self.unit.async_getValues(*a,**kw)

    def async_setValues(self,*a,**kw):
        return self.unit.async_setValues(*a,**kw)

    async def load(self, path: str | None = None, data: dict | None = None):
        await super().load(path, data)
        self.data = fixup(self.cfg, root=self.cfg, path=Path(), this_file=self.cfg_path)
        await self.add_registers()

    async def add_registers(self):
        """Replace entries w/ register/slot members with Register instances"""

        async def a_r(d, path=Path()):
            seen = False
            for k, v in d.items():
                if not isinstance(v, dict):
                    continue
                if await a_r(v, path / k):
                    seen = True
                    if "register" in v:
                        logger.warning("%s has a sub-register: ignored", path / k)
                    elif "slot" in v:
                        logger.warning("%s is not a register", path / k)
                    continue

                if "register" in v:
                    d[k] = reg = self.factory(v, path / k)
                    await reg.start()
                    self.unit.add(get_kind(v.get("reg_type")), offset=v.register, val=reg)
                    seen = True
            return seen

        await a_r(self.cfg)



"""
implement a Modbus server.

Its job is to forward requests to clients, multiplexing the Modbus
connection to them, because some of these things are stupid and
only support one concurrent TCP connection.

"""

from __future__ import annotations

import logging

import anyio
from pymodbus.exceptions import ModbusIOException
from pymodbus.factory import ServerDecoder
from pymodbus.pdu import ExceptionResponse
from pymodbus.transaction import ModbusSocketFramer

logger = logging.getLogger(__name__)

__all__ = ["Server"]


class Forwarder:
    "Handler for one connection"

    def __init__(self, server, client):
        self.server = server
        self.client = client
        self.sem = anyio.Semaphore(10)
        self.framer = ModbusSocketFramer(ServerDecoder())
        self.send_lock = anyio.Lock()

    async def work(self, request):
        """Subtask to process a request, then send the reply"""
        unit_id = request.unit_id
        transaction_id = request.transaction_id
        try:
            try:
                unit = self.server.units[unit_id]
            except KeyError:
                response = ExceptionResponse(
                    request.function_code, ExceptionResponse.SLAVE_FAILURE
                )
            else:
                response = await unit.process_request(request)

            response.unit_id = unit_id
            response.transaction_id = transaction_id
            response = self.framer.buildFrame(response)
            try:
                async with self.send_lock:
                    await self.client.send(response)
            except (
                ConnectionResetError,
                anyio.ClosedResourceError,
                anyio.BrokenResourceError,
                anyio.EndOfStream,
            ):
                pass

        finally:
            self.sem.release()

    async def run(self):
        """Client task. Reads requests and starts handlers to
        process them asynchronously."""

        async with self.client, anyio.create_task_group() as tg:
            try:
                data = bytearray()
                while True:
                    data += await self.client.receive(4096)

                    msgs = []

                    while True:
                        used, pdu = self.framer.processIncomingFrame(data)
                        data = data[used:]
                        if pdu is None:
                            break
                        msgs.append(pdu)

                    for msg in msgs:
                        await self.sem.acquire()
                        tg.start_soon(self.work, msg)

            except (
                anyio.IncompleteRead,
                ConnectionRefusedError,
                ConnectionResetError,
                anyio.ClosedResourceError,
                ModbusIOException,
                anyio.BrokenResourceError,
                anyio.EndOfStream,
                TimeoutError,
            ):
                tg.cancel_scope.cancel()
                return


class Server:
    """
    A forwarding Modbus server.
    """

    def __init__(self, host=None, port=502):
        self.host = host
        self.port = port
        self.units = {}

    def attach(self, unit: int, dev):
        """Attach a Modbus device under this unit#"""
        self.units[unit] = dev

    async def run(self):
        """Run the server."""

        async def proc(client):
            fwd = Forwarder(self, client)
            await fwd.run()

        listener = await anyio.create_tcp_listener(local_host=self.host, local_port=self.port)
        await listener.serve(proc)



"""
Various types
"""

from __future__ import annotations

import struct

import anyio

from pymodbus.pdu.register_message import (
    ReadHoldingRegistersRequest,
    ReadHoldingRegistersResponse,
    ReadInputRegistersRequest,
    ReadInputRegistersResponse,
    WriteMultipleRegistersRequest,
    WriteMultipleRegistersResponse,
    WriteSingleRegisterRequest,
    WriteSingleRegisterResponse,
)
from pymodbus.pdu.bit_message import (
    ReadCoilsRequest,
    ReadCoilsResponse,
    ReadDiscreteInputsRequest,
    ReadDiscreteInputsResponse,
    WriteSingleCoilRequest,
    WriteSingleCoilResponse,
    WriteMultipleCoilsRequest,
    WriteMultipleCoilsResponse,
)

MAX_REQ_LEN = 30

import logging

from pymodbus.datastore.store import BaseModbusDataBlock

logger = logging.getLogger(__name__)


def _singleton(x):
    x = x()
    return x


class BaseValue:
    """Base class for a single value.

    Do not instantiate directly.

    Clear @idem if unchanged items shall trigger iterators.
    """

    len = 0
    _value = None
    gen = 0
    block: DataBlock = None
    to_write: int = None

    def __init__(self, value=None, *, offset=None, idem=True):
        self.changed = anyio.Event()
        self._value = value
        self._value_w = value
        self.idem = idem
        self.offset = offset

        if value is not None:
            self.gen = 1
            self.changed = anyio.Event()

    @property
    def value_w(self):
        "value to be written to the bus"
        return self._value_w

    @property
    def value(self):
        "value read from the bus"
        return self._value

    @value.setter
    def value(self, val: int | float):
        "sets the value that's read from the bus"
        self._value = self._constrain(val)

    def set(self, val: int | float, idem: bool = False):
        """Set the value-to-be-written.

        Triggers a write unless @idem is set (default: it is not).
        """
        # pylint: disable=missing-function-docstring
        if idem and self._value == val:
            return
        val = self._constrain(val)
        self._value_w = val
        self.gen += 1
        self.to_write = self.gen
        if self.block is not None:
            self.block.trigger_send()

    def _constrain(self, val):
        return val

    def _decode(self, regs):
        raise NotImplementedError

    def _encode(self, value):
        raise NotImplementedError

    def decode(self, regs: list[int]) -> None:
        """
        Decode the passed-in register value(s) into this variable.

        Triggers iterators, except if the value doesn't change
        and ``self.idem`` is set.
        """
        val = self._decode(regs)
        if self.idem and val == self.value:
            return
        self.value = self._decode(regs)
        self.changed.set()
        self.changed = anyio.Event()
        self.gen += 1

    def clear(self) -> None:
        """
        Clears the value.
        """
        if self.value is None:
            return
        self.value = None
        self.gen += 1
        self.changed.set()
        self.changed = anyio.Event()

    def encode(self) -> list[int]:
        """
        Encode the current value. Returns a list of registers.
        """
        return self._encode(self.value_w)

    def __str__(self):
        return f"‹{self.value}›"

    def __repr__(self):
        res = f"<{self.__class__.__name__}"
        if self.offset is not None:
            res += f" @{self.offset}"
        res += f":{self._value}"
        if self._value_w != self._value:
            res += f":{self._value_w}"
        res += ">"

        return res

    def __aiter__(self):
        return ValueIterator(self)


class _Signed:
    """A mix-in for signed integers."""

    # pylint: disable=no-member,abstract-method

    def _decode(self, regs):
        res = super()._decode(regs)
        if res is not None and res & (1 << (self.len * 16 - 1)):
            res -= 1 << (self.len * 16)
        return res

    def _encode(self, value):
        if value < 0:
            value += 1 << (self.len * 16)
        return super()._encode(value)


class _Swapped:
    """A mix-in to byteswap the Modbus data."""

    # pylint: disable=no-member,abstract-method

    def _decode(self, regs):
        regs = list(regs)
        regs.reverse()
        return super()._decode(regs)

    def _encode(self, value):
        regs = list(super()._encode(value))
        regs.reverse()
        return tuple(regs)


class InaccessibleValue(BaseValue):  # duck-types but does NOT interit BaseValue
    """This register range must not be accessed.

    Use an instance of this type (with appropriate length)
    to force splitting a request into multiple parts.

    :param len: The length of the block that may not be accessed.
    """

    # pylint: disable=abstract-method

    def __init__(self, length):
        # pylint: disable=super-init-not-called
        self.len = length


class IntValue(BaseValue):
    """Simplest-possible value, one register."""

    len = 1

    def _constrain(self, val):
        if val is None:
            return val
        return int(val)

    def _decode(self, regs):
        return regs[0]

    def _encode(self, value):
        return (value,)


class LongValue(IntValue):
    """32-bit integer, two registers, standard (big-endian) word order."""

    len = 2

    def _decode(self, regs):
        return (regs[0] << 16) | regs[1]

    def _encode(self, value):
        return (value >> 16, value & 0xFFFF)


class QuadValue(IntValue):
    """64-bit integer, four registers, standard (big-endian) word order.

    This is a BaseValue instance.
    """

    len = 4

    def _decode(self, regs):
        return (((((regs[0] << 16) + regs[1]) << 16) + regs[2]) << 16) + regs[3]

    def _encode(self, value):
        return (
            (value >> 48) & 0xFFFF,
            (value >> 32) & 0xFFFF,
            (value >> 16) & 0xFFFF,
            value & 0xFFFF,
        )


class BitValue(BaseValue):
    """Bits, for "coil"s."""

    len = 1

    def _decode(self, regs):
        return bool(regs[0])

    def _encode(self, value):
        return (int(value),)


class InvBitValue(BaseValue):
    """Inverted bits, for "coil"s."""

    len = 1

    def _decode(self, regs):
        return not bool(regs[0])

    def _encode(self, value):
        return (int(not value),)


class FloatValue(BaseValue):
    """network-ordered floating point."""

    len = 2

    def _decode(self, regs):
        return struct.unpack(">f", struct.pack(">2H", *regs))[0]

    def _encode(self, value):
        return struct.unpack(">2H", struct.pack(">f", value))


class DoubleValue(BaseValue):
    """network-ordered accurate floating point."""

    len = 4

    def _decode(self, regs):
        return struct.unpack(">d", struct.pack(">4H", *regs))[0]

    def _encode(self, value):
        return struct.unpack(">4H", struct.pack(">d", value))


class SwappedLongValue(_Swapped, LongValue):
    """32-bit integer, two registers, little-endian word order."""

    pass


class SwappedQuadValue(_Swapped, QuadValue):
    """64-bit integer, four registers, little-endian word order."""

    pass


class SignedIntValue(_Signed, IntValue):
    """one register, signed."""

    pass


class SignedLongValue(_Signed, LongValue):
    """two registers, signed."""

    pass


class SwappedSignedLongValue(_Signed, _Swapped, LongValue):
    """two registers, signed, swapped."""

    pass


SignedSwappedLongValue = SwappedSignedLongValue


class SignedQuadValue(_Signed, QuadValue):
    """four registers, signed."""

    pass


class SwappedSignedQuadValue(_Signed, _Swapped, QuadValue):
    """four registers, signed, swapped."""

    pass


SignedSwappedQuadValue = SwappedSignedQuadValue


class SwappedFloatValue(_Swapped, FloatValue):
    """broken-ordered floating point."""

    pass


class SwappedDoubleValue(_Swapped, DoubleValue):
    """broken-ordered accurate floating point."""

    pass


class ByteValue(BaseValue):
    """Bytestring. @length is in bytes"""

    endian = ">"
    len = -1

    def __init__(self, length, *a, **kw):
        self.len = (length + 1) // 2
        self.pack = f"{self.endian}{self.len}H"

        super().__init__(*a, **kw)

    def _encode(self, value):
        ln = self.len * 2
        value += b"\0" * (ln - len(value))

        return struct.unpack(self.pack, value)

    def _decode(self, regs):
        return struct.pack(self.pack, *regs)


class SwappedByteValue(ByteValue):
    """Bytes, null-terminated, little-endian words. @length is in bytes"""

    endian = "<"


class StringValue(ByteValue):
    """Text, null-terminated.
    @length is in bytes, NOT UTF-8 characters"""

    def _encode(self, value):
        value = value.encode("utf-8")
        return super()._encode(value)

    def _decode(self, regs):
        return super()._decode(regs).rstrip(b"\0").decode("utf-8")


class SwappedStringValue(StringValue):
    """Text, null-terminated, little-endian words.
    @length is in bytes, NOT UTF-8 characters"""

    endian = "<"


class TypeCodec:
    """Base class for access types. Do not instantiate."""

    typ = None
    acc = None

    def __repr__(self):
        return self.__class__.__name__

    def __eq__(self, typ):
        if isinstance(typ, TypeCodec):
            typ = typ.typ
        return self.typ == typ

    def __hash__(self):  # pylint: disable=invalid-hash-returned
        return self.typ


@_singleton
class Coils(TypeCodec):
    """Modbus 'coils' data.
    This is a TypeCodec.
    """

    typ = 0
    key = "c"
    encoder = ReadCoilsRequest
    decoder = ReadCoilsResponse
    encoder_s = WriteSingleCoilRequest
    decoder_s = WriteSingleCoilResponse
    encoder_m = WriteMultipleCoilsRequest
    decoder_m = WriteMultipleCoilsResponse


@_singleton
class DiscreteInputs(TypeCodec):
    """Modbus 'discrete input' data.
    This is a TypeCodec.
    """

    typ = 1
    key = "d"
    encoder = ReadDiscreteInputsRequest
    decoder = ReadDiscreteInputsResponse


@_singleton
class HoldingRegisters(TypeCodec):
    """Modbus 'holding register' data.
    This is a TypeCodec.
    """

    typ = 2
    key = "h"
    encoder = ReadHoldingRegistersRequest
    decoder = ReadHoldingRegistersResponse
    encoder_s = WriteSingleRegisterRequest
    decoder_s = WriteSingleRegisterResponse
    encoder_m = WriteMultipleRegistersRequest
    decoder_m = WriteMultipleRegistersResponse


@_singleton
class InputRegisters(TypeCodec):
    """Modbus 'input register' data.
    This is a TypeCodec.
    """

    typ = 3
    key = "i"
    encoder = ReadInputRegistersRequest
    decoder = ReadInputRegistersResponse
    encoder_s = WriteSingleRegisterRequest
    decoder_s = WriteSingleRegisterResponse
    encoder_m = WriteMultipleRegistersRequest
    decoder_m = WriteMultipleRegistersResponse


class DataBlock(dict, BaseModbusDataBlock):
    """Your basic sparse data block.

    The @changed attribute is an event that triggers when a write request
    succeeds.
    """

    def __init__(self, max_rd_len=MAX_REQ_LEN, max_wr_len=MAX_REQ_LEN):
        super().__init__()
        self.max_rd_len = max_rd_len
        self.max_wr_len = max_wr_len
        self.changed = anyio.Event()

    def __bool__(self):
        return True

    def reset(self):
        """
        Clear all values in this block
        """
        for val in self.values():
            val.value = None

    def trigger_send(self):
        """Called by a value when it's been changed.

        Currently a no-op in this class.
        """
        pass

    def add(self, offset: int, val: BaseValue):
        """Add a value to the block."""
        if offset in self:
            raise ValueError("Already known", offset)
        for n in range(1, 8):
            try:
                if self[offset - n].len > n:
                    raise ValueError(f"Overlap with {self[offset - n]} @{offset - n}")
                break
            except KeyError:
                pass
        for n in range(1, val.len):
            try:
                if offset + n in self:
                    raise ValueError(f"Overlap with {self[offset + n]} @{offset + n}")
                break
            except KeyError:
                pass
        self[offset] = val
        val.block = self

    def validate(self, address: int, count: int = 1):
        "does nothing. Compatibility with pymodbus 3.8"
        return True

    def ranges(self, changed=False, max_len=MAX_REQ_LEN):
        """Iterate over to-be-retrieved/sent range(s).

        If @changed is set, skip unmodified items.
        """
        start, cur = None, None
        for offset, val in sorted(self.items()):
            if isinstance(val, InaccessibleValue):
                if start is not None:
                    yield (start, cur - start)
                    start = None
            elif changed and val.to_write is None:
                continue
            elif start is None:
                start = offset
                cur = start + val.len
            elif cur == offset and (cur + val.len - start) <= max_len:
                cur += val.len
            else:
                yield (start, cur - start)
                start = offset
                cur = start + val.len

        if cur is not None:
            yield (start, cur - start)

    def getValues(self, address: int, count=1) -> list[int]:
        """Returns the array of Modbus values for the @address:+@count range

        Called when preparing a Send request.
        """
        res = []
        address -= 1
        while count > 0:
            try:
                val = self[address]
            except KeyError:
                res.append(0)
                address += 1
                count -= 1
            else:
                try:
                    res.extend(val.encode())
                except TypeError as exc:
                    raise RuntimeError(f"Cannot encode {val!r}") from exc
                address += val.len
                count -= val.len
        if count < 0:
            # well, this shouldn't happen but …
            res = res[:count]
        return res

    def markSent(self, address: int, count=1):
        """Mark this range as sent: remember gen counter"""
        while count > 0:
            try:
                val = self[address]
            except KeyError:
                address += 1
                count -= 1
            else:
                if val.to_write is not None:
                    if val.to_write == val.gen:
                        val.to_write = None
                    else:
                        val.to_write = val.gen
                        self.trigger_send()

                address += val.len
                count -= val.len

    def setValues(self, address: int, values: list[int]):
        """Set the variables starting at @address to @values.

        Called with the reply of a Read request.
        """
        address -= 1
        while values:
            try:
                val = self[address]
            except KeyError:
                address += 1
                values.pop(0)
            else:
                val.decode(values[: val.len])
                address += val.len
                values = values[val.len :]

        self.changed.set()
        self.changed = anyio.Event()

    def delete(self, address, count=1):
        """
        Remove @count values at @address
        """
        while count:
            val = self.pop(address, None)
            if val is not None:
                val.block = None
            address += val.len
            count -= 1


class ValueIterator:
    """
    Helper class for iterating over `BaseValue` changes.

    Posts a notification when values get skipped.
    """

    def __init__(self, val):
        self.val = val
        self.gen = max(0, val.gen - 1)

    async def __anext__(self):
        """
        Iterate over values / value changes.

        If the value is initially unknown, wait.
        if it's been cleared, raises `StopAsyncIteration`.
        """
        val = self.val

        if val.gen > 0 and val.value is None:
            raise StopAsyncIteration
        if self.gen == val.gen:
            await val.changed.wait()
        if val.value is None:
            raise StopAsyncIteration
        if self.gen + 1 != val.gen:
            logger.info("%r: skipped %d", val, val.gen - self.gen - 1)
        self.gen = val.gen
        return val.value



#!/usr/bin/env python3
"""
Basic "modbus" tool: network client and server, serial client

"""

from __future__ import annotations

import logging  # pylint: disable=wrong-import-position
from getopt import getopt
from pprint import pprint

import asyncclick as click
import anyio

from .typemap import get_type, map_kind, map_type
from .types import InputRegisters, IntValue

log = logging.getLogger()


@click.group()
async def main():
    """Modbus client / server"""

    FORMAT = (
        "%(asctime)-15s %(threadName)-15s %(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s"
    )
    logging.basicConfig(format=FORMAT)
    log.setLevel(logging.WARNING)


UNIT = 0x1


def _doc(v):
    return v.__doc__.split("\n")[0]


_args_kind = "\n".join(f"{k:3s}  {_doc(v)}" for k, v in map_kind.items())
_args_type = "\n".join(f"{k:3s}  {_doc(v)}" for k, v in map_type.items())
_args_help = f"""\
\b
Setting values:
-u UNIT   register to this unit; default 1
-k KIND   use this register range, default Input
-t TYPE   register this datatype, default Raw
-r REGNUM start at this register, default 0, auto-incremented
-n COUNT  register this many variables, default 1
-v VALUE  set the register(s) to this value

\b
Kinds:
{_args_kind}

\b
Types:
{_args_type}
"""


async def _server(host, port, debug, args):
    """
    Basic Modbus server, for static tests.
    """
    if debug:
        log.setLevel(logging.DEBUG)
    if not args:
        raise click.UsageError("You didn't add any values to serve")

    from moat.modbus.server import ModbusServer  # pylint: disable=import-outside-toplevel

    s = ModbusServer(address=host, port=port)
    unit = None
    kind = InputRegisters
    typ = IntValue
    num = 1
    reg = 0

    kv, a = getopt(
        args,
        "u:k:t:r:n:v:",
        "--unit= --kind= --type= reg= register= num= val= value=".split(),
    )
    if a:
        raise click.UsageError(f"Unknown argument: {' '.join(a)}")
    for k, v in kv:
        pend = True
        if k in ("-u", "--unit"):
            unit = s.add_unit(int(v))
        elif k in ("-k", "--kind"):
            kind = map_kind[v[0]]
        elif k in ("-t", "--type"):
            typ = get_type(v[0])
        elif k in ("-r", "--reg", "--register"):
            reg = int(v)
        elif k in ("-n", "--num"):
            num = int(v)
        elif k in ("-v", "--val", "--value"):
            try:
                v = int(v)
            except ValueError:
                v = float(v)
            if unit is None:
                unit = s.add_unit(1)
            for _ in range(num):
                unit.add(kind, reg, typ(v))
                reg += typ.len
            pend = False
        else:
            raise click.UsageError(f"Unknown argument: {k!r}")
    if pend:
        raise click.UsageError("Values must be at the end")

    await s.serve()


def mk_server(m):
    """Helper to create a server"""
    s = _server
    s = click.argument("args", nargs=-1, type=click.UNPROCESSED)(s)
    s = click.option("--debug", "-d", is_flag=True, help="Log debug messages")(s)
    s = click.option("--port", "-p", type=int, default=502, help="port to bind to")(s)
    s = click.option("--host", "-s", default="localhost", help="host to bind to")(s)
    s = m.command(
        "server",
        context_settings=dict(
            show_default=True,
            ignore_unknown_options=True,
            help_option_names=["-?", "--help"],
        ),
        epilog=_args_help,
    )(s)
    return s


server = mk_server(main)


def flint(v):
    """float-or-int"""
    try:
        return int(v)
    except ValueError:
        return float(v)


async def _client(host, port, unit, kind, start, num, type_, values, debug, interval, maxlen):
    """
    Basic Modbus-TCP client.
    """
    if debug:
        log.setLevel(logging.DEBUG)

    from moat.modbus.client import ModbusClient  # pylint: disable=import-outside-toplevel

    c = {}
    if maxlen:
        c["max_rd_len"] = maxlen
        c["max_wr_len"] = maxlen
    async with (
        ModbusClient() as g,
        g.host(host, port, **c) as h,
        h.unit(unit) as u,
        u.slot("default") as s,
    ):
        k = map_kind[kind[0]]
        t = get_type(type_)
        if values:
            if len(values) == 1:
                values = values * num
            elif num == 1:
                num = len(values)
            elif len(values) != num:
                raise click.UsageError("One or N values!")
        for i in range(num):
            v = s.add(k, start, t)
            if values:
                v.set(flint(values[i]), idem=False)
            start += t.len
            num -= 1

        last_res = None
        while True:
            if values:
                await s.setValues()  # pylint:disable=protected-access  ## TODO
            else:
                res = await s.getValues()  # pylint:disable=protected-access  ## TODO
                if last_res != res:
                    pprint(res)
                    last_res = res

            if interval:
                await anyio.sleep(interval)
            else:
                break


def mk_client(m):
    """helper to create a client"""
    c = _client
    c = click.argument("values", nargs=-1)(c)
    c = click.option("--debug", "-d", is_flag=True, help="Log debug messages")(c)
    c = click.option(
        "--type",
        "-t",
        "type_",
        default="raw",
        help="value type: raw, s1,u2,f4, c#/b#:text/bytes, x/X:bool/inv; Sx/Fx=swapped",
    )(c)
    c = click.option("--num", "-n", type=int, default=1, help="number of values")(c)
    c = click.option("--start", "-s", default=0, help="starting register")(c)
    c = click.option("--kind", "-k", default="i", help="query type: input, discrete, hold, coil")(
        c,
    )
    c = click.option("--unit", "-u", type=int, default=1, help="unit to query")(c)
    c = click.option("--port", "-p", type=int, default=502, help="destination port")(c)
    c = click.option("--host", "-h", default="localhost", help="destination host")(c)
    c = click.option("--interval", "-i", type=float, help="loop: read/write every N seconds")(c)
    c = click.option(
        "--max-len",
        "-L",
        "maxlen",
        type=int,
        default=30,
        help="max. modbus words per packet",
    )(c)
    c = m.command("client", context_settings=dict(show_default=True))(c)
    return c


client = mk_client(main)


async def _serclient(
    port,
    baudrate,
    stopbits,
    parity,
    unit,
    kind,
    start,
    num,
    type_,
    values,
    debug,
    maxlen,
):
    """
    Basic Modbus-RTU client.
    """
    if debug:
        log.setLevel(logging.DEBUG)

    from moat.modbus.client import ModbusClient  # pylint: disable=import-outside-toplevel

    c = {}
    if maxlen:
        c["max_rd_len"] = maxlen
        c["max_wr_len"] = maxlen
    async with (
        ModbusClient() as g,
        g.serial(port=port, baudrate=baudrate, stopbits=stopbits, parity=parity, **c) as h,
        h.unit(unit) as u,
        u.slot("default") as s,
    ):
        k = map_kind[kind[0]]
        t = get_type(type_)
        if values:
            if len(values) == 1:
                values = values * num
            elif len(values) != num:
                raise click.UsageError("One or N values!")
        for i in range(num):
            v = s.add(k, start, t)
            if values:
                v.set(flint(values[i]))
            start += t.len
            num -= 1

        try:
            if values:
                await s.setValues()  # pylint:disable=protected-access  ## TODO
            else:
                res = await s.getValues()  # pylint:disable=protected-access  ## TODO
                pprint(res)
        except Exception as exc:  # pylint: disable=broad-except
            log.exception("Problem: %r", exc)


def add_serial_cfg(c):
    """Helper for serial port configuration"""
    c = click.option(
        "--port",
        "-p",
        required=True,
        type=str,
        help="destination port (/dev/ttyXXX)",
    )(c)
    c = click.option("--baudrate", "-b", type=int, default=9600, help="Baud rate (9600)")(c)
    c = click.option("--parity", "-P", type=str, default="N", help="Parity (NEO), default N")(c)
    c = click.option("--stopbits", "-S", type=int, default=1, help="Stopbits (12), default 1")(c)
    return c


def mk_serial_client(m):
    """helper to create a sserial client"""
    c = _serclient
    c = click.argument("values", nargs=-1)(c)
    c = click.option("--debug", "-d", is_flag=True, help="Log debug messages")(c)
    c = click.option(
        "--type",
        "-t",
        "type_",
        default="raw",
        help="value type: s1,s2,s4,u1,u2,u4,f2,f4,raw; Sx/Fx=swapped",
    )(c)
    c = click.option("--num", "-n", type=int, default=1, help="number of values")(c)
    c = click.option("--start", "-s", default=0, help="starting register")(c)
    c = click.option("--kind", "-k", default="i", help="query type: input, discrete, hold, coil")(
        c,
    )
    c = add_serial_cfg(c)
    c = click.option("--unit", "-u", type=int, default=1, help="unit to query")(c)
    c = click.option(
        "--max-len",
        "-L",
        "maxlen",
        type=int,
        default=30,
        help="max. modbus words per packet",
    )(c)
    c = m.command("serial", context_settings=dict(show_default=True))(c)
    return c


serialclient = mk_serial_client(main)


if __name__ == "__main__":
    main(_anyio_backend="trio")  # pylint: disable=unexpected-keyword-arg



"""
Support values on MoaT-KV
"""

from __future__ import annotations

import logging

from .device import Register as BaseRegister
from moat.util import Path, NotGiven
import anyio

logger = logging.getLogger(__name__)


class Register(BaseRegister):
    """
    One possibly-complex Modbus register that's mirrored from and/or to MoaT-KV
    """

    def __init__(self, *a, link=None, tg=None, is_server=False, **kw):
        super().__init__(*a, **kw)
        self._link = link
        self.tg = tg
        self.is_server = is_server

        if self.src is None and self.dest is None:
            if self.data.get("slot", "write") != "write":
                if getattr(self.data, "_is_orig", False):
                    logger.warning("%s:%s: no source/destination", self.unit, self.path)
            return

    async def start(self):
        await super().start()

        # logger.info("%s:%s: Polling", self.unit, self.path)
        tg = self.tg

        if (dest := self.dest) is not None:
            if self.is_server:
                slot = None
            else:
                slot = self.data.get("slot", None)
                if slot is None:
                    logger.warning("%s:%s: no read slot", self.unit, self.path)

            # logger.info("%s:%s: Write %s", self.unit, self.path, dest)
            if isinstance(dest, Path):
                dest = (dest,)

            for d in dest:
                tg.start_soon(self.to_link, d)

        if self.src is not None:
            slot = self.data.get("slot", None) if self.dest is None else None
            # if a slot is set AND src is set AND dst is not set,
            # then we want to do a periodic write (keepalive etc.).
            mon = self._link.d_watch(self.src, meta=True)

            # logger.info("%s:%s: Watch %s", self.unit, self.path,self.src)

            if self.is_server or slot in (None, "write"):
                await tg.start(self.from_link, mon)
            else:
                tg.start_soon(self.from_link_p, mon, self.slot.write_delay)

    @property
    def src(self):
        return self.data.get("src")

    @property
    def dest(self):
        return self.data.get("dest")

    def set(self, val):
        self.reg.set(val)

    async def to_link(self, dest):
        """Copy a Modbus value to MoaT-Link"""
        async for val in self:
            logger.debug("%s L %r", self.path, val)
            await self._link.d_set(dest, val, retain=True)

    async def from_link(self, mon, *, task_status):
        """Copy an MQTT value to Modbus"""
        async with mon as mon_:
            if task_status is not None:
                task_status.started()
                task_status = None
            async for val in mon_:
                if val is None:  # Link message
                    continue
                if isinstance(val,tuple):  # Link client
                    logger.debug("%s W %r", self.path, val)
                    val=val[0]
                elif "value" not in val:  # KV message
                    logger.debug("%s Wx", self.path)
                    continue
                else:
                    logger.debug("%s W %r", self.path, val)
                    val=val.value
                await self._set(val)

    async def from_link_p(self, mon, slot):
        """Copy an MQTT value to Modbus, with periodic refresh"""
        evt = anyio.Event()
        val = NotGiven

        async def per():
            nonlocal evt, val
            while True:
                if val is NotGiven:
                    await evt.wait()
                else:
                    with anyio.move_on_after(slot):
                        await evt.wait()
                if val is NotGiven:
                    continue

                logger.debug("%s Wr %r", self.path, val)
                await self._set(val)

        async with mon as mon_, anyio.create_task_group() as tg:
            tg.start_soon(per)
            first = True
            async for val_ in mon_:
                if isinstance(val,tuple):
                    val = val[0]
                else:
                    try:
                        val = val_.value
                    except AttributeError:
                        if first:
                            first = False
                            continue
                        val = NotGiven
                logger.debug("%s w %r", self.path, val)
                evt.set()
                evt = anyio.Event()

    async def _set(self, value):
        self.value = value

        if (dest := self.dest) is not None and self.data.get("mirror", False):
            if isinstance(dest, Path):
                dest = (dest,)

            for d in dest:
                await self._link.d_set(d, value, retain=True)



"""
Device file handling, with a basic multi-device client
"""

from __future__ import annotations

import logging
from pathlib import Path as FSPath

import asyncclick as click
from moat.util import yload, yprint

from .device import fixup
from .poll import dev_poll

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """Modbus device polling"""
    pass


@cli.command()
@click.option("-r", "--raw", is_flag=True, help="don't postprocess")
@click.option("-R", "--no-refs", is_flag=True, help="don't process references")
@click.argument("path", type=click.Path("r"))
def dump(path, raw, no_refs):
    """Dump a postprocessed file"""
    path = FSPath(path)
    with path.open("r") as f:
        d = yload(f, attr=True)
    if not raw:
        d = fixup(d, do_refs=not no_refs, this_file=path)
    yprint(d)


@cli.command()
@click.argument("path", type=click.File("r"))
@click.pass_context
async def poll(ctx, path):
    """Poll Modbus devices as directed via YAML."""
    obj = ctx.obj

    cfg = yload(path, attr=True)

    if "link" in obj.cfg:
        # pylint: disable=import-outside-toplevel
        from moat.link.client import Link

        ln_ctx = Link(opj.cfg.link)
    else:
        ln_ctx = nullcontext(None)

    async with ln_ctx as mt_ln:
        await dev_poll(cfg, link=mt_ln)



#!/usr/bin/python3

"""
Modbus server classes for serial(RTU) and TCP.
"""

from __future__ import annotations

import logging
import socket
from binascii import b2a_hex
from contextlib import asynccontextmanager
import time

import anyio
from anyio.abc import SocketAttribute
from moat.util import CtxObj

try:
    from pymodbus.datastore import ModbusServerContext, ModbusDeviceContext
except ImportError:
    from pymodbus.datastore import ModbusServerContext
    from pymodbus.datastore import ModbusSlaveContext as ModbusDeviceContext
from pymodbus.device import ModbusControlBlock, ModbusDeviceIdentification
from pymodbus.exceptions import NoSuchSlaveException
from pymodbus.pdu import ExceptionResponse, DecodePDU
from pymodbus.utilities import hexlify_packets

from moat.modbus.types import BaseValue, DataBlock, TypeCodec

_logger = logging.getLogger(__name__)


class UnitContext(ModbusDeviceContext):
    """
    This module implements a slave context for servers that stores
    individual variables.
    """

    def __init__(self, server=None, unit=None):
        super().__init__(
            di=DataBlock(),
            co=DataBlock(),
            ir=DataBlock(),
            hr=DataBlock(),
        )
        if server:
            self.unit = unit
            server._add_unit(self)

    def add(
        self,
        typ: TypeCodec,
        offset: int,
        val: BaseValue | type[BaseValue],
    ) -> BaseValue:
        """Add a field to be served.

        :param typ: The `TypeCodec` instance to use.
        :param offset: The value's numeric offset, zero-based.
        :param val: The data type (baseValue instance)

        `cls` is either the decoder (subclass of `BaseValue`),
        or an existing `BaseValue` instance.
        """
        k = self.store[typ.key]
        if isinstance(val, type):
            val = val()
        k.add(offset, val)
        return val

    def remove(self, typ: TypeCodec, offset: int):
        """Remove a field to be requested.

        :param typ: the `TypeCodec` to use
        :param offset: the offset where the value is located

        Returns the field in question, or none if it doesn't exist.
        """
        k = self.store[typ.key]
        return k.delete(offset + 1)


class BaseModbusServer(CtxObj):
    """Basic base class for servers."""

    def __init__(self, identity=None, response_manipulator=None):
        self.context = ModbusServerContext(single=False)
        self.context._slaves = self.units = {}
        self.control = ModbusControlBlock()
        self.broadcast_enable = False
        self.response_manipulator = response_manipulator

        if isinstance(identity, ModbusDeviceIdentification):
            self.control.Identity.update(identity)

        if identity is None:
            identity = ModbusDeviceIdentification()
            identity.VendorName = "Matthias Urlichs"
            identity.ProductCode = "MoaT.modbus"
            identity.VendorUrl = "http://M-o-a-T.org/"
            identity.ProductName = "MoaT-Modbus Test"
            identity.ModelName = "MoaT-Modbus Test"
            identity.MajorMinorRevision = "1.0"
        self.identity = identity

    def add_unit(self, unit, ctx=None) -> UnitContext:
        """
        Add an empty unit (= slave context) to this server (and return it).

        The unit must not exist.
        """
        if unit in self.units:
            raise RuntimeError(f"Unit {unit} already exists")
        if ctx is None:
            return UnitContext(self, unit)
        self.units[unit] = ctx

    def _add_unit(self, unit):
        self.units[unit.unit] = unit

    async def serve(self, opened=None):
        """The actual server. Override me."""
        raise RuntimeError("You need to override .serve")

    async def process_request(self, request):
        """Basic request processor"""
        context = self.context[request.dev_id]
        if hasattr(context, "process_request"):
            response = await context.process_request(request)
        else:
            response = await request.update_datastore(context)
        return response

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as tg:
            evt = anyio.Event()
            tg.start_soon(self.serve, evt)
            try:
                await evt.wait()
                yield self
            finally:
                tg.cancel_scope.cancel()


class SerialModbusServer(BaseModbusServer):
    """
    A simple serial Modbus server (RTU).
    """

    _serial = None
    framer = None
    ignore_missing_slaves = False
    single = False

    def __init__(self, identity=None, timeout=None, **args):
        super().__init__(identity=identity)
        self.args = args
        self.timeout = timeout

        from pymodbus.framer.rtu_framer import (  # pylint: disable=import-outside-toplevel
            ModbusRtuFramer,
        )

        class Framer(ModbusRtuFramer):
            def _validate_dev_id(self, unit, single):
                return True

        self.decoder = DecodePDU(False)  # pylint: disable=no-value-for-parameter ## duh?
        self.Framer = Framer

    async def serve(self, opened=None):
        from anyio_serial import Serial  # pylint: disable=import-outside-toplevel

        async with Serial(**self.args) as ser:
            self._serial = ser
            self.framer = self.Framer(self.decoder)

            if opened is not None:
                opened.set()
            t = time.monotonic()
            while True:
                with anyio.move_on_after(0.1):
                    await ser.receive()
                    break
            while True:
                if self.timeout:
                    with anyio.fail_after(self.timeout):
                        data = await ser.receive()
                else:
                    data = await ser.receive()
                t2 = time.monotonic()
                if t2 - t > 0.2:
                    self.framer.resetFrame()
                t = t2
                msgs = []
                while True:
                    used, pdu = self.framer.processIncomingFrame(data)
                    data = data[used:]
                    if pdu is None:
                        break
                    msgs.append(pdu)

                for msg in msgs:
                    with anyio.fail_after(2):
                        await self._process(msg)

    async def _process(self, request):
        broadcast = False
        unit = request.unit_id
        tid = request.transaction_id

        try:
            if self.broadcast_enable and not request.unit_id:
                broadcast = True
                # if broadcasting then execute on all slave contexts,
                # note response will be ignored
                for unit_id in self.context.slaves():
                    response = await request.execute(self.context[unit_id])
            else:
                response = await self.process_request(request)
        except NoSuchSlaveException:
            txt = f"requested slave does not exist: {request.unit_id}"
            _logger.error(txt)
            if self.ignore_missing_slaves:
                return  # the client will simply timeout waiting for a response
            response = ExceptionResponse(
                request.function_code, ExceptionResponse.GATEWAY_NO_RESPONSE
            )
        except Exception:  # pylint: disable=broad-except
            _logger.exception("Unable to fulfill request")
            response = ExceptionResponse(request.function_code, ExceptionResponse.SLAVE_FAILURE)
        # no response when broadcasting
        response.unit_id = unit
        response.transaction_id = tid
        if isinstance(response, ExceptionResponse):
            _logger.error(
                "Source: %r %d %d %d",
                type(request),
                request.function_code,
                request.address,
                getattr(request, "count", 1),
            )

        if not broadcast:
            response.transaction_id = request.transaction_id
            response.unit_id = request.unit_id
            skip_encoding = False
            if self.response_manipulator:
                response, skip_encoding = self.response_manipulator(response)
            if not skip_encoding:
                response = self.framer.buildFrame(response)
            #           if _logger.isEnabledFor(logging.DEBUG):
            #               _logger.debug("send: [%s]- %s", request, b2a_hex(response))

            await self._serial.send(response)


def create_server(cfg):
    """
    Create a server (TCP or serial) according to the configuration.

    @cfg is a dict with either host/port or port/serial keys.
    """
    if "serial" in cfg:
        port = cfg.get("port", None)
        kw = cfg["serial"]
        if port is not None:
            kw["port"] = port
        return SerialModbusServer(**kw)
    elif "host" in cfg or ("port" in cfg and isinstance(cfg["port"], int)):
        kw = {}
        for k, v in cfg.items():
            if isinstance(k, str):
                if k == "units":
                    continue
                if k == "host":
                    k = "address"
                kw[k] = v
        return ModbusServer(**kw)
    else:
        raise ValueError("neither serial nor TCP config found")


class RelayServer:
    """
    A mix-in class to teach a server to forward all requests to a client
    """

    single = True

    def __init__(self, client, *a, **k):
        self._client = client
        super().__init__(*a, **k)

    async def _process(self, request):
        request = self.mon_request(request)
        tid = request.transaction_id
        resp = await self._client.execute(request)

        resp.transaction_id = tid
        resp = self.mon_response(resp) or resp
        resp = self.framer.buildFrame(resp)  # pylint:disable=no-member
        await self._serial.send(resp)  # pylint:disable=no-member

    def mon_request(self, request):
        """Request monitor. Override me."""
        return request

    def mon_response(self, response):
        """Response monitor. Override me."""
        return response


class ModbusServer(BaseModbusServer):
    """TCP Modbus server.

    If the identity structure is not passed in, the ModbusControlBlock
    uses its own empty structure.

    :param identity: An optional identity structure
    :param address: An optional address to bind to.
    :param port: the TCP port to listen on.
    """

    taskgroup = None
    single = False

    def __init__(self, identity=None, address=None, port=None):
        super().__init__(identity=identity)

        from pymodbus.framer.socket import (  # pylint: disable=import-outside-toplevel
            FramerSocket,
        )

        self.decoder = DecodePDU(True)
        self.framer = FramerSocket
        self.address = address or "localhost"
        self.port = port if port is not None else 502

    async def serve(self, opened: anyio.Event | None = None):
        """Run this server.
        Sets the `opened` event, if given, as soon as the server port is open.
        """
        try:
            async with anyio.create_task_group() as tg:
                self.taskgroup = tg
                async with await anyio.create_tcp_listener(
                    local_port=self.port,
                    local_host=self.address,
                    reuse_port=True,
                ) as server:
                    if self.port == 0:
                        self.port = server.listeners[0].extra_attributes[
                            SocketAttribute.local_address
                        ]()[1]
                    if opened is not None:
                        opened.set()

                    await server.serve(self._serve_one)
        except socket.gaierror:
            _logger.error("Trying to look up %s", self.address)
            raise
        finally:
            self.taskgroup = None

    async def _serve_one(self, conn):
        reset_frame = False
        framer = self.framer(decoder=self.decoder)

        while True:
            try:
                data = await conn.receive(4096)
                if data == b"":
                    break
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(  # pylint: disable=logging-not-lazy
                        "Handling data: " + hexlify_packets(data),
                    )

                reqs = []
                while True:
                    used, pdu = framer.processIncomingFrame(data)
                    data = data[used:]
                    if pdu is None:
                        break
                    reqs.append(pdu)

                for request in reqs:
                    unit = request.dev_id
                    tid = request.transaction_id
                    try:
                        response = await self.process_request(request)
                    except NoSuchSlaveException:
                        _logger.debug("requested slave does not exist: %d", request.dev_id)
                        response = ExceptionResponse(
                            request.function_code, ExceptionResponse.GATEWAY_NO_RESPONSE
                        )
                    except Exception as exc:  # pylint: disable=broad-except
                        _logger.warning("Datastore unable to fulfill request", exc_info=exc)
                        response = ExceptionResponse(
                            request.function_code, ExceptionResponse.SLAVE_FAILURE
                        )
                    response.transaction_id = tid
                    response.dev_id = unit
                    # self.server.control.Counter.BusMessage += 1
                    pdu = framer.buildFrame(response)
                    if _logger.isEnabledFor(logging.DEBUG):
                        _logger.debug("send: %s", b2a_hex(pdu))
                    await conn.send(pdu)

            except TimeoutError as msg:
                _logger.debug("Socket timeout occurred: %r", msg)
                reset_frame = True
            except OSError as msg:
                _logger.error("Socket error occurred: %r", msg)
                return
            except anyio.get_cancelled_exc_class():
                raise
            except anyio.BrokenResourceError:
                return
            except Exception:  # pylint: disable=broad-except
                _logger.exception("Server error")
                return
            finally:
                if reset_frame:
                    framer.resetFrame()
                    reset_frame = False


class MockAioModbusServer(ModbusServer):
    """A test modbus server with static data"""

    pass



"""
Basic "moat modbus" tool: network client and server, serial client
"""

from __future__ import annotations

import sys
import traceback

import anyio
import asyncclick as click
from moat.util import load_subgroup

from moat.modbus.client import ModbusClient
from moat.modbus.server import RelayServer, SerialModbusServer

from pymodbus.pdu.register_message import WriteSingleRegisterRequest

from .__main__ import add_serial_cfg, mk_client, mk_serial_client, mk_server


@load_subgroup(sub_pre="moat.modbus")
async def cli():
    """Modbus tools"""
    pass


serialclient = mk_serial_client(cli)

client = mk_client(cli)
server = mk_server(cli)


def print_exc(exc, **kw):  # pylint: disable=missing-function-docstring
    traceback.print_exception(type(exc), exc, exc.__traceback__, **kw)


@cli.group(invoke_without_command=True)
@add_serial_cfg
@click.option("-t", "--timeout", type=float, default=0, help="Error if no more data (seconds)")
@click.option(
    "-T",
    "--initial-timeout",
    "timeout1",
    type=float,
    default=0,
    help="Error if no data (seconds)",
)
@click.pass_context
async def monitor(ctx, timeout, timeout1, **params):
    """
    A basic Modbus RTU monitor.

    This command dumps the messages on a Modbus RTU line.

    The arguments address the link to the server, i.e. it acts as a Modbus client.
    You need to add a "to" subcommand with the connection that should act as
    the server, i.e. which the actual client is connected to.

    Otherwise this is TODO, as distinguishing client
    and server messages is somewhat nontrivial.

    """
    obj = ctx.obj

    if ctx.invoked_subcommand is not None:
        obj.A = params
        obj.timeout = timeout
        obj.timeout1 = timeout1
        return

    raise click.UsageError("Single line monitoring is not implemented yet")


#   async def mon(msg):
#       print(msg)

#   async with ModbusClient() as g, g.serial(monitor=mon, **params) as h:
#       g, h  # pylint: disable=pointless-statement
#       if obj.debug > 1:
#           print("Listening.", file=sys.stderr)
#       while True:
#           await anyio.sleep(99999)


@monitor.command
@add_serial_cfg
@click.option("-r", "--retry", type=int, help="Delay between restarts in case of errors")
@click.pass_obj
async def to(obj, retry, **params):
    """
    This subcommand describes the ModBus interface of the client(s).

    Useful for reverse engineering, serial speed/format translation, debugging …
    """
    A = None
    B = None

    class Server(RelayServer, SerialModbusServer):
        """A time-out-ing serial Modbus relay"""

        def __init__(self, *a, **kw):
            self.__evt = anyio.Event()
            super().__init__(*a, **kw)

        def mon_request(self, request):
            "request monitor"
            if isinstance(request, WriteSingleRegisterRequest):
                print(f"> {request}", request.value)
            else:
                print(f"> {request}", getattr(request, "registers", ""))
            return request

        def mon_response(self, response):
            "response monitor"
            print(f"< {response}", getattr(response, "registers", ""))
            self.__evt.set()
            return response

        async def watch(self, t2, t1):
            "Timeout manager"
            t = t1
            while True:
                if t is None:
                    await self.__evt.wait()
                else:
                    with anyio.fail_after(t):
                        await self.__evt.wait()
                t = t2
                self.__evt = anyio.Event()

    while True:
        try:
            async with (
                ModbusClient() as g_a,
                g_a.serial(**obj.A) as A,
                Server(client=A, **params) as B,
                # anyio.create_task_group() as tg,
            ):
                await B.watch(obj.timeout, obj.timeout1)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            if not retry or not A or not B:
                raise
            print_exc(exc, file=sys.stderr)
            print(f"Retrying in {retry}s", file=sys.stderr)
            await anyio.sleep(retry)



"""
Poll code
"""

from __future__ import annotations

import logging
from functools import partial

import anyio
from asyncscope import scope
from moat.util import attrdict, merge, to_attrdict

from moat.modbus.client import ModbusClient
from .device import ServerDevice, ClientDevice, fixup
from moat.modbus.server import create_server

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from moat.link.client import Link

logger = logging.getLogger(__name__)


async def dev_poll(cfg:dict, link:Link, *, task_status=anyio.TASK_STATUS_IGNORED):
    """
    Run a device task on this set of devices, as configured by the config.

    The config will be preprocessed by `moat.modbus.dev.device.fixup`; the
    result will be returned via @task_status after setup is complete.
    """
    sl = cfg.setdefault("slots", attrdict())
    sl.setdefault("write", attrdict())
    sl._apply_default = True  # pylint:disable=protected-access
    cfg = fixup(cfg)

    s = cfg.setdefault("src", attrdict())
    sl = cfg.slots

    async with ModbusClient() as cl, anyio.create_task_group() as tg:
        nd = 0

        async def make_dev(v, Reg, **kw):
            kw = to_attrdict(kw)
            vs = v.setdefault("src", attrdict())
            merge(vs, kw, replace=False)
            vsl = v.setdefault("slots", attrdict())
            merge(vsl, sl, replace=False)

            logger.info("Starting %r", vs)

            dev = ClientDevice(client=cl, factory=Reg)
            await dev.load(data=v)

            # return await scope.spawn_service(dev.as_scope)
            async def task(dev, *, task_status):
                async with dev:
                    task_status.started(dev)
                    await anyio.sleep_forever()
            return await tg.start(task, dev)

        if link is None:
            from .device import Register as Reg  # pylint: disable=import-outside-toplevel

            RegS = Reg
        else:
            # The MoaT-Link client must live longer than the taskgroup
            from .link import Register  # pylint: disable=import-outside-toplevel

            Reg = partial(Register, link=link, tg=tg)
            RegS = partial(Register, link=link, tg=tg, is_server=True)

        # relay-out server(s)
        servers = []
        for s in cfg.get("server", ()):
            srv = create_server(s)
            servers.append(srv)

            for u, v in s.get("units", {}).items():
                dev = ServerDevice(factory=RegS)
                await dev.load(data=v)
                srv.add_unit(u, dev)
                nd += 1

        def do_attach(v, dev):
            p = v.get("server", None)
            if p is None:
                return
            if isinstance(p, int):
                s, u = servers[p // 1000], p % 1000
            else:
                s, u = servers[p[0]], p[1]
            s.add_unit(u, dev.unit)

            nonlocal nd
            nd += 1

        # serial clients
        for h, hv in cfg.get("ports", {}).items():
            try:
                sp = hv["serial"]
            except KeyError:
                logger.error("No serial params for port %r", h)
                continue
            for u, v in hv.items():
                if not isinstance(u, int):
                    continue
                dev = await make_dev(v, Reg, port=h, serial=sp, unit=u)
                tg.start_soon(dev.poll)
                do_attach(v, dev)

        # TCP clients
        for h, hv in cfg.get("hosts", {}).items():
            for u, v in hv.items():
                if not isinstance(u, int):
                    continue
                dev = await make_dev(v, Reg, host=h, unit=u)
                tg.start_soon(dev.poll)
                do_attach(v, dev)

        # more TCP clients
        for h, hv in cfg.get("hostports", {}).items():
            for p, pv in hv.items():
                if not isinstance(p, int):
                    continue
                for u, v in pv.items():
                    dev = await make_dev(v, Reg, host=h, port=p, unit=u)
                    tg.start_soon(dev.poll)
                    do_attach(v, dev)

        for s in servers:
            evt = anyio.Event()
            tg.start_soon(partial(s.serve, opened=evt))
            await evt.wait()

        task_status.started(cfg)

        if nd:
            logger.info("Running.")
        if not nd:
            logger.error("No devices to poll found.")

        pass  # wait until all tasks are done
